package com.github.polpetta.mezzotre.orm;

import com.github.polpetta.mezzotre.orm.model.CallbackQueryContext;
import com.github.polpetta.mezzotre.orm.model.query.QCallbackQueryContext;
import com.google.errorprone.annotations.CanIgnoreReturnValue;
import io.ebean.typequery.PString;
import java.util.concurrent.CompletableFuture;
import java.util.function.Supplier;
import javax.inject.Inject;
import javax.inject.Singleton;
import org.slf4j.Logger;

/**
 * This class allows the system to lazily remove {@link CallbackQueryContextCleaner} entries that
 * are no more needed, using the {@link BatchBeanCleanerService} service under the hood.
 *
 * @author Davide Polonio
 * @see BatchBeanCleanerService
 * @since 1.0
 */
@Singleton
public class CallbackQueryContextCleaner {

  private static final Supplier<PString<QCallbackQueryContext>> ENTRY_GROUP =
      () -> new QCallbackQueryContext().entryGroup;
  private static final Supplier<PString<QCallbackQueryContext>> SINGLE_ENTRY =
      () -> new QCallbackQueryContext().id;

  private final BatchBeanCleanerService batchBeanCleanerService;
  private final Logger log;

  @Inject
  public CallbackQueryContextCleaner(BatchBeanCleanerService batchBeanCleanerService, Logger log) {
    this.batchBeanCleanerService = batchBeanCleanerService;
    this.log = log;
  }

  /**
   * Add a group of {@link com.github.polpetta.mezzotre.orm.model.CallbackQueryContext} to be
   * removed
   *
   * @param id the group id of the {@link
   *     com.github.polpetta.mezzotre.orm.model.CallbackQueryContext}
   * @return a {@link CompletableFuture} with an {@link Integer} indicating how many entries have
   *     been deleted from the persistence layer
   * @see CallbackQueryContext#getEntryGroup()
   */
  @CanIgnoreReturnValue
  public CompletableFuture<Integer> removeGroupAsync(String id) {
    log.trace("CallbackQueryContext entry group " + id + " queued for removal");
    return batchBeanCleanerService.removeAsync(id, ENTRY_GROUP.get());
  }

  /**
   * Add a single {@link com.github.polpetta.mezzotre.orm.model.CallbackQueryContext} to be removed
   *
   * @param id the id of the {@link CallbackQueryContext} to remove
   * @return a {@link CompletableFuture} with an {@link Integer} indicating how many entries have
   *     been deleted from the persistence layer. Can be 0 or 1.
   * @see CallbackQueryContext#getId()
   */
  @CanIgnoreReturnValue
  public CompletableFuture<Integer> removeIdAsync(String id) {
    log.trace("CallbackQueryContext single entity " + id + " queued for removal");
    return batchBeanCleanerService.removeAsync(id, SINGLE_ENTRY.get());
  }
}
