package com.github.polpetta.mezzotre.telegram.command;

import static org.junit.jupiter.api.Assertions.*;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.when;

import com.github.polpetta.mezzotre.helper.Loader;
import com.github.polpetta.mezzotre.helper.TestConfig;
import com.github.polpetta.mezzotre.i18n.LocalizedMessageFactory;
import com.github.polpetta.mezzotre.i18n.TemplateContentGenerator;
import com.github.polpetta.mezzotre.orm.model.Campaign;
import com.github.polpetta.mezzotre.orm.model.TgChat;
import com.github.polpetta.mezzotre.orm.model.query.QCampaign;
import com.github.polpetta.mezzotre.orm.model.query.QTgChat;
import com.github.polpetta.mezzotre.orm.telegram.CampaignUtil;
import com.github.polpetta.mezzotre.orm.telegram.ChatUtil;
import com.github.polpetta.mezzotre.util.Clock;
import com.github.polpetta.mezzotre.util.UUIDGenerator;
import com.github.polpetta.types.json.ChatContext;
import com.google.gson.Gson;
import com.pengrad.telegrambot.model.Update;
import com.pengrad.telegrambot.request.BaseRequest;
import com.pengrad.telegrambot.request.SendMessage;
import io.ebean.Database;
import java.util.Objects;
import java.util.Optional;
import java.util.concurrent.CompletableFuture;
import java.util.concurrent.Executors;
import java.util.concurrent.TimeUnit;
import java.util.stream.Collectors;
import java.util.stream.Stream;
import org.apache.velocity.app.VelocityEngine;
import org.junit.jupiter.api.*;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.Arguments;
import org.junit.jupiter.params.provider.MethodSource;
import org.testcontainers.containers.PostgreSQLContainer;
import org.testcontainers.junit.jupiter.Container;
import org.testcontainers.junit.jupiter.Testcontainers;

@Tag("slow")
@Tag("database")
@Tag("velocity")
@Testcontainers
class CreateCampaignIntegrationTest {

  private static Gson gson;

  @Container
  private final PostgreSQLContainer<?> postgresServer =
      new PostgreSQLContainer<>(TestConfig.POSTGRES_DOCKER_IMAGE);

  private VelocityEngine velocityEngine;
  private Database database;
  private UUIDGenerator fakeUUIDGenerator;
  private Clock fakeClock;
  private ChatUtil chatUtil;
  private CampaignUtil campaignUtil;
  private CreateCampaign createCampaign;

  @BeforeAll
  static void beforeAll() {
    gson = new Gson();
  }

  @BeforeEach
  void setUp() throws Exception {
    database =
        Loader.connectToDatabase(Loader.loadDefaultEbeanConfigWithPostgresSettings(postgresServer));
    velocityEngine = Loader.defaultVelocityEngine();

    fakeClock = mock(Clock.class);
    fakeUUIDGenerator = mock(UUIDGenerator.class);

    chatUtil = new ChatUtil(fakeClock);
    campaignUtil = new CampaignUtil(fakeUUIDGenerator);

    createCampaign =
        new CreateCampaign(
            new TemplateContentGenerator(new LocalizedMessageFactory(velocityEngine)),
            Executors.newSingleThreadExecutor(),
            fakeUUIDGenerator,
            chatUtil,
            campaignUtil);
  }

  static Stream<Arguments> getValidCampaignEntries() {
    return Stream.of(
        Arguments.of("a simple name", "a simple description"),
        Arguments.of("a simple name", "a simple\nmultiline\n description"),
        Arguments.of(
            "an \uD83E\uDD21 emoji \uD83D\uDCAF name \uD83D\uDD25",
            "a very \uD83E\uDDE8 campaign description \uD83E\uDD16"),
        Arguments.of("一个简单的名字", "一个简单的描述"));
  }

  @ParameterizedTest
  @MethodSource("getValidCampaignEntries")
  @Timeout(value = 1, unit = TimeUnit.MINUTES)
  void shouldCreateANewCampaign(String name, String description) throws Exception {
    when(fakeUUIDGenerator.generateAsString()).thenReturn("64b483b4-7de8-4632-9117-927756d26210");
    final TgChat chat =
        new TgChat(
            42L, new ChatContext(), "en-US", true // doesn't really matter
            );
    chat.save();

    {
      final Update firstMessage =
          gson.fromJson(
              "{\n"
                  + "\"update_id\":10000,\n"
                  + "\"message\":{\n"
                  + "  \"date\":1441645532,\n"
                  + "  \"chat\":{\n"
                  + "     \"last_name\":\"Test Lastname\",\n"
                  + "     \"id\":42,\n"
                  + "     \"type\": \"private\",\n"
                  + "     \"first_name\":\"Test Firstname\",\n"
                  + "     \"username\":\"Testusername\"\n"
                  + "  },\n"
                  + "  \"message_id\":1365,\n"
                  + "  \"from\":{\n"
                  + "     \"last_name\":\"Test Lastname\",\n"
                  + "     \"id\":1111111,\n"
                  + "     \"first_name\":\"Test Firstname\",\n"
                  + "     \"username\":\"Testusername\"\n"
                  + "  },\n"
                  + "  \"text\":\"/createCampaign\"\n"
                  + "}\n"
                  + "}",
              Update.class);

      final CompletableFuture<Optional<BaseRequest<?, ?>>> got =
          createCampaign.process(chat, firstMessage);
      final Optional<BaseRequest<?, ?>> baseRequestOptional = got.get();
      final BaseRequest<?, ?> gotResponse = baseRequestOptional.get();
      assertInstanceOf(SendMessage.class, gotResponse);
      assertEquals(
          "Very well, give me a second that I go grab some scroll, ink and quill...\n"
              + "\n"
              + "_*Searches frenetically in the desk between piles of books and written scrolls for"
              + " a free piece of scroll*_\n"
              + "\n"
              + "Here we are! Please, tell me, how would you like to call your new campaign?",
          gotResponse.getParameters().get("text"));

      final ChatContext retrievedChatContext =
          Objects.requireNonNull(new QTgChat().id.eq(42L).findOne()).getChatContext();
      assertEquals(1, retrievedChatContext.getStep());
      assertEquals("/createCampaign", retrievedChatContext.getStage());
    }

    {
      final Update secondMessageName =
          gson.fromJson(
              "{\n"
                  + "\"update_id\":10000,\n"
                  + "\"message\":{\n"
                  + "  \"date\":1441645532,\n"
                  + "  \"chat\":{\n"
                  + "     \"last_name\":\"Test Lastname\",\n"
                  + "     \"id\":1111111,\n"
                  + "     \"type\": \"private\",\n"
                  + "     \"first_name\":\"Test Firstname\",\n"
                  + "     \"username\":\"Testusername\"\n"
                  + "  },\n"
                  + "  \"message_id\":1365,\n"
                  + "  \"from\":{\n"
                  + "     \"last_name\":\"Test Lastname\",\n"
                  + "     \"id\":1111111,\n"
                  + "     \"first_name\":\"Test Firstname\",\n"
                  + "     \"username\":\"Testusername\"\n"
                  + "  },\n"
                  + "  \"text\":\""
                  + name
                  + "\"\n"
                  + "}\n"
                  + "}",
              Update.class);

      final CompletableFuture<Optional<BaseRequest<?, ?>>> got =
          createCampaign.process(chat, secondMessageName);
      final Optional<BaseRequest<?, ?>> baseRequestOptional = got.get();
      final BaseRequest<?, ?> gotResponse = baseRequestOptional.get();
      assertInstanceOf(SendMessage.class, gotResponse);
      assertEquals(
          "Uhmm yes, "
              + name
              + "... Very well, great name choice by the way! Now, would you like to give me a"
              + " short description of your campaign? You can skip this step if you don't want,"
              + " otherwise just reply with the description",
          gotResponse.getParameters().get("text"));

      final ChatContext retrievedChatContext =
          Objects.requireNonNull(new QTgChat().id.eq(42L).findOne()).getChatContext();
      assertEquals(2, retrievedChatContext.getStep());
      assertEquals("/createCampaign", retrievedChatContext.getStage());
      assertEquals(name, retrievedChatContext.getAdditionalProperties().get("campaign_name"));
    }

    {
      final Update thirdMessageDescription =
          gson.fromJson(
              "{\n"
                  + "\"update_id\":10000,\n"
                  + "\"message\":{\n"
                  + "  \"date\":1441645532,\n"
                  + "  \"chat\":{\n"
                  + "     \"last_name\":\"Test Lastname\",\n"
                  + "     \"id\":1111111,\n"
                  + "     \"type\": \"private\",\n"
                  + "     \"first_name\":\"Test Firstname\",\n"
                  + "     \"username\":\"Testusername\"\n"
                  + "  },\n"
                  + "  \"message_id\":1365,\n"
                  + "  \"from\":{\n"
                  + "     \"last_name\":\"Test Lastname\",\n"
                  + "     \"id\":1111111,\n"
                  + "     \"first_name\":\"Test Firstname\",\n"
                  + "     \"username\":\"Testusername\"\n"
                  + "  },\n"
                  + "  \"text\":\""
                  + description
                  + "\"\n"
                  + "}\n"
                  + "}",
              Update.class);

      final CompletableFuture<Optional<BaseRequest<?, ?>>> got =
          createCampaign.process(chat, thirdMessageDescription);
      final Optional<BaseRequest<?, ?>> baseRequestOptional = got.get();
      final BaseRequest<?, ?> gotResponse = baseRequestOptional.get();
      assertInstanceOf(SendMessage.class, gotResponse);
      assertEquals(
          "_*Finishes to write down on the scroll the last details*_\n"
              + "\n"
              + "All done! I have successfully added your campaign to my archives! Let me store it"
              + " properly...\n"
              + "\n"
              + "_*Moves the scroll into a pile of other, disorganized, scrolls*_\n"
              + "\n"
              + "Now, if you want, you can share this campaign to other players so that they can"
              + " join! To do this, just click on the button below and then select the user or the"
              + " group you want to share this campaign with! They will be able to add their"
              + " characters and see the other characters in this campaign",
          gotResponse.getParameters().get("text"));

      final ChatContext retrievedChatContext =
          Objects.requireNonNull(new QTgChat().id.eq(42L).findOne()).getChatContext();
      assertEquals(3, retrievedChatContext.getStep());
      assertEquals("/createCampaign", retrievedChatContext.getStage());
      assertEquals(
          0,
          retrievedChatContext.getAdditionalProperties().size(),
          () ->
              "Additional Properties should be empty, instead got: \n"
                  + retrievedChatContext.getAdditionalProperties().entrySet().stream()
                      .map(e -> "key: " + e.getKey() + " -> value: " + e.getValue())
                      .collect(Collectors.joining("\n")));
    }

    final Campaign gotCampaign =
        new QCampaign().name.eq(name).id.eq("64b483b4-7de8-4632-9117-927756d26210").findOne();
    assertNotNull(gotCampaign);
    // First two assertions kinda pointless btw
    assertEquals("64b483b4-7de8-4632-9117-927756d26210", gotCampaign.getId());
    assertEquals(name, gotCampaign.getName());
    assertEquals(description, gotCampaign.getDescription());
  }

  // TODO add invalid name tests, add InlineButtonTests too
}
